PROGRAM SkinTut2;

{**********************************************************}
{                                                          }
{   Skins program tutorial 2                               }
{                                                          }
{   Copyright (c) 2002 by Leon de Boer                     }
{   ldeboer@attglobal.net  - primary e-mail address        }
{   ldeboer@iprimus.com.au - backup e-mail address         }
{                                                          }
{****************[ THIS CODE IS FREEWARE ]*****************}
{                                                          }
{     This sourcecode is released for the purpose to       }
{   promote the pascal language on all platforms. You may  }
{   redistribute it and/or modify with the following       }
{   DISCLAIMER.                                            }
{                                                          }
{     This SOURCE CODE is distributed "AS IS" WITHOUT      }
{   WARRANTIES AS TO PERFORMANCE OF MERCHANTABILITY OR     }
{   ANY OTHER WARRANTIES WHETHER EXPRESSED OR IMPLIED.     }
{                                                          }
{*****************[ SUPPORTED PLATFORMS ]******************}
{                                                          }
{     FreePascal 1.0.6+     Windows 32 Bit                 }
{                                                          }
{*****************[ REVISION HISTORY ]*********************}
{  Version  Date        Fix                                }
{  -------  ---------   ---------------------------------  }
{  1.00     9 May 02    First release                      }
{**********************************************************}
{  Ok now on a real skins program the skin is an external  }
{  file containing the images to skin. For this tutorial   }
{  we will move the skin to an external bitmap file called }
{  skintut2.bmp which you may feel free to alter and when  }
{  you run the program the image will change. To create a  }
{  hole in the bitmap you paint the area to remove in      }
{  magenta (Red=255,Green=0,Blue=255). You can replace     }
{  the whole bitmap with a new one if you like. The load   }
{  of a bitmap from an external file should be a simple    }
{  call to LoadImage but as always with Microsoft we had   }
{  bugs with this call in Windows NT and XP. Hence we have }
{  to have a bit of code to fix the LoadImage call.        }
{                                                          }
{  If you look at the code line marked as "Ref1" you can   }
{  see this line determines what colour the outline is     }
{  created from. You can change the colour to another if   }
{  you like. If you want to know a colours RGB in a bitmap }
{  go into "Paint" and use the eyedropper on the bitmap.   }
{  You can also set a range of colours by increasing the   }
{  tolerance value out.                                    }
{                                                          }
{**********************************************************}

{$APPTYPE GUI}                                        { FULL GUI MODE }
USES Windows;


CONST
   szClassName    = 'SkinDemo';                       { The class name }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{  LoadBmpFromFile -> Load bitmap from file          Updated -> 9May02 LdB  }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{  This is a pascal port of a C function workaround for the LoadImage API   }
{  call by Microsoft (Q158898). The function reads a BMP file and returns   }
{  a hanle to the loaded bitmap and the palette for the bitmap.             }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
FUNCTION LoadBmpFromFile (FileName: String; VAR Pal: HPalette): HBitmap;
VAR I: Integer; szFileName: String; Bmp, OldBmp: HBitMap; Bm: TBitMap;
    MemDc: hDc; RGB: Array [0..255] Of TRGBQuad; LogPal: PLogPalette;
BEGIN
   szFileName := FileName + #0;                       { Make asciiz name }
   Bmp := LoadImage(HInstance, @szFilename[1],
     IMAGE_BITMAP, 0, 0, LR_DEFAULTCOLOR OR
     LR_CREATEDIBSECTION OR LR_LOADFROMFILE);         { Load bitmap }
   IF (Bmp <> 0) THEN BEGIN
     GetObject(Bmp, SizeOf(Bm), @Bm);                 { Get bitmap data }
     IF ((Bm.bmBitsPixel * Bm.bmPlanes) <= 8) THEN
     BEGIN                                            { 256 colours or less }
       MemDc := CreateCompatibleDc(0);                { Create memory DC }
       OldBmp := SelectObject(MemDc, Bmp);            { Select this bitmap }
       GetDIBColorTable(MemDc, 0, 256, RGB);          { Get color table }
       GetMem(LogPal, SizeOf(TLogPalette) +
         (256*SizeOf(TPaletteEntry)));                { Allocate a memory }
       LogPal^.palVersion := $0300;                   { Set palette version }
       LogPal^.palNumEntries := 256;                  { Set entry count }
       FOR I := 0 TO 255 DO
         WITH LogPal^ DO BEGIN
           palPalEntry[I].peRed := RGB[I].rgbRed;     { Transfer red colour }
           palPalEntry[I].peGreen := RGB[I].rgbGreen; { Transfer green colour }
           palPalEntry[I].peBlue := RGB[I].rgbBlue;   { Transfer blue colour }
           palPalEntry[I].peFlags := 0;               { Clear flag entry }
         END;
       Pal := CreatePalette(LogPal^);                 { Create a palette }
       FreeMem(LogPal, SizeOf(TLogPalette) +
         (256*SizeOf(TPaletteEntry)));                { Release memory }
       SelectObject(MemDc, OldBmp);                   { Restore old bitmap }
       DeleteObject(MemDc);                           { Release context }
     END ELSE BEGIN
       MemDc := GetDc(0);                             { Get ref context }
       Pal := CreateHalftonePalette(MemDc);           { Create palette }
       ReleaseDc(0, MemDc);                           { Release context }
     END;
   END;
   LoadBmpFromFile := Bmp;                            { Return the bitmap }
END;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{  BitmapToRegion -> Create region from bitmap       Updated -> 8May02 LdB  }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{  This is a pascal port of a C function by Jean-Edouard Lachand-Robert     }
{  (http://www.geocities.com/Paris/LeftBank/1160/resume.htm), June 1998.    }
{  It takes a bitmap and creates a region based upon the pixels in the      }
{  bitmap matching a color that corresponds to Transcolor with a            }
{  Color Tolerance level of Tolerance.                                      }
{  It does this by scanning each line in the BITMAP structure for           }
{  pixels that match TransColor then adds them to the region with a         }
{  call to CombineRgn.                                                      }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
FUNCTION BitmapToRegion(BmpName: String; TransColor: TColorRef; Tolerance: Byte): HRGN;
CONST ALLOC_UNIT = 100;
VAR B, LR, LG, LB, HR, HG, HB: Byte; X, X0, Y: LongInt; MaxRects: DWord;
    pbits32: Pointer; MemDC, TempDc: HDc; BitmapInfo: TBitmapInfo; Pal: HPalette;
    InBmp, hbm32: HBitmap; Bm, bm32: TBITMAP; hData: HGLOBAL; pData: PRgnData;
    p32: pByte; p: pLongInt; pr: PRect; h: HRGN; HoldObj, HoldObj1: HGDIObj;
    OldPal: HPalette; Rslt: HRGN;
BEGIN
  Rslt := 0;                                          { Preset fail result }
  InBmp := LoadBmpFromFile(BmpName, Pal);             { Load main bitmap }
  IF (InBmp <> 0) THEN BEGIN                          { Bitmap loaded }
    MemDC := CreateCompatibleDC(0);                   { Create memory context }
    IF (MemDC <> 0) THEN BEGIN                        { Memory context valid }
      GetObject(InBmp, SizeOf(Bm), @Bm);              { Get bitmap data }
      WITH BitmapInfo.bmiHeader DO BEGIN              { Make bitmap header }
        biSize          := SizeOf(TBitmapInfoHeader); { Set record size }
        biWidth         := Bm.bmWidth;                { Width of bitmap }
        biHeight        := Bm.bmHeight;               { Height of bitmap }
        biPlanes        := 1;                         { 1 plane }
        biBitCount      := 32;                        { 32 bit DIB }
        biCompression   := BI_RGB;                    { uncompressed format }
        biSizeImage     := 0;                         { Default size }
        biXPelsPerMeter := 0;                         { Default x res }
        biYPelsPerMeter := 0;                         { Default y res }
        biClrUsed       := 0;                         { No unused colours }
        biClrImportant  := 0;                         { No important colours }
      END;
      hbm32 := CreateDIBSection(MemDC, BitmapInfo,
        DIB_RGB_COLORS, pbits32, 0, 0);               { Create a DIB section }
      IF (hbm32 <> 0) THEN BEGIN
        HoldObj := SelectObject(MemDC, hbm32);        { Select DIB section }
        TempDC := CreateCompatibleDC(0);              { Create memory DC }
        HoldObj1 := SelectObject(TempDC, InBmp);      { Select bitmap }
        OldPal := SelectPalette(TempDc, Pal, True);   { Select palette }
        RealizePalette(TempDc);                       { Realize the palette }
        BitBlt(MemDC, 0, 0, Bm.bmWidth, Bm.bmHeight,
          TempDc, 0, 0, SRCCOPY);                     { Transfer bitmap }
        SelectPalette(TempDc, OldPal, False);         { Restore old palette }
        SelectObject(TempDc, HoldObj1);               { Select original object }
        DeleteObject(TempDc);                         { Delete memory context }
        GetObject(hbm32, SizeOf(bm32), @bm32);        { Fetch bitmap data }
        WHILE (bm32.bmWidthBytes MOD 4) > 0 DO
          Inc(bm32.bmWidthBytes);                     { Make divisable by 4 }
        MaxRects := ALLOC_UNIT;                       { Set maximum size }
        hData := GlobalAlloc(GMEM_MOVEABLE, SizeOf(TRgnDataHeader) +
          (SizeOf(TRect) * MaxRects));                { Allocate memory }
        pData := GlobalLock(hData);                   { Lock the memory }
        pData^.rdh.dwSize := SizeOf(TRgnDataHeader);  { Set memory size }
        pData^.rdh.iType := RDH_RECTANGLES;           { Rectangular area }
        pData^.rdh.nCount := 0;                       { Zero count }
        pData^.rdh.nRgnSize := 0;                     { Zero region size }
        SetRect(pData^.rdh.rcBound, MaxInt, MaxInt,
          0, 0);                                      { Make rect maximum }
        LR := GetRValue(TransColor);                  { Lowest red }
        LG := GetGValue(TransColor);                  { Lowest green }
        LB := GetBValue(TransColor);                  { Lowest blue }
        If (LR + Tolerance > $FF) Then HR := $FF      { Check not above max }
          ELSE HR := LR + Tolerance;                  { New high red value }
        If (LG + Tolerance > $FF) Then HG := $FF      { Check not above max }
          ELSE HG := LG + Tolerance;                  { New high green value }
        If (LB + Tolerance > $FF) Then HB := $FF      { Check not above max }
          ELSE HB := LB + Tolerance;                  { New high blue value }
        p32 := bm32.bmBits;                           { Pointer to bmp data }
        Inc(PChar(p32), LongInt(bm32.bmHeight - 1) *
          LongInt(bm32.bmWidthBytes));                { Move to bottom line }
        FOR Y := 0 to Bm.bmHeight-1 DO BEGIN          { Each line of bitmap }
          X := -1;                                    { Start left }
          WHILE (X+1 < Bm.bmWidth) DO BEGIN           { For each pixel }
            Inc(X);                                   { Next x position }
            X0 := X;                                  { Hold left x value }
            p := PLongInt(p32);                       { Hold pointer }
            Inc(PChar(p), X * SizeOf(LongInt));       { Set p start value }
            WHILE (X < Bm.bmWidth) DO BEGIN           { For each pixel }
              B := GetBValue(p^);                     { Get blue value }
              IF (B >= LR) AND (B <= HR) THEN BEGIN   { Check for match }
                B := GetGValue(p^);                   { Get green value }
                IF (B >= LG) AND (B <= HG) THEN BEGIN
                  B := GetRValue(p^);                 { Get red value }
                  IF (B >= LB) AND (B <= HB) THEN     { Pixel is "transparent" }
                    Break;                            { Just exit }
                END;
              END;
              Inc(PChar(P), SizeOf(LongInt));         { Advance pointer }
              Inc(X);                                 { Next x position }
            END;
            IF (X > X0) THEN BEGIN                    { New region to add }
              IF (pData^.rdh.nCount >= MaxRects)
              THEN BEGIN                              { Area is to big }
                GlobalUnlock(hData);                  { Unlock memory }
                Inc(MaxRects, ALLOC_UNIT);            { Adjust size }
                hData := GlobalReAlloc(hData, SizeOf(TRgnDataHeader) +
                  SizeOf(TRect) * MaxRects, GMEM_MOVEABLE);
                pData := GlobalLock(hData);           { Relock memory }
              END;
              pr := @pData^.Buffer[pData^.rdh.nCount * SizeOf(TRect)];
              SetRect(Pr, X0, Y, X, Y+1);             { Set rectangle coords }
              IF (X0 < pData^.rdh.rcBound.Left) THEN
                pData^.rdh.rcBound.Left := X0;        { Check left x value }
              IF (Y < pData^.rdh.rcBound.Top) THEN
                pData^.rdh.rcBound.Top := Y;          { Check top y value }
              IF (X > pData^.rdh.rcBound.Right) THEN
                pData^.rdh.rcBound.Left := X;         { Check right x value }
              IF (Y+1 > pData^.rdh.rcBound.Bottom) THEN
                pData^.rdh.rcBound.Bottom := Y+1;     { Check bottom y value }
              Inc(pData^.rdh.nCount);                 { Add 1 region }
              IF (pData^.rdh.nCount = 2000)           { Reach maximum region count }
              THEN BEGIN                              { Region count at max }
                h := ExtCreateRegion(NIL, SizeOf(TRgnDataHeader) +
                  (SizeOf(TRect)*MaxRects), pData^);  { Create a region }
                IF (Rslt <> 0) THEN BEGIN             { Already have region }
                  CombineRgn(Rslt, Rslt, h, RGN_OR);
                  DeleteObject(h);                    { Delete old region }
                END ELSE Rslt := h;                   { Hold region }
                pData^.rdh.nCount := 0;               { Rezero region count }
                SetRect(pData^.rdh.rcBound, MaxInt,
                  MaxInt, 0, 0);                      { Reset rectangle }
              END;
            END;
          END;
          Dec(PChar(p32), bm32.bmWidthBytes);         { Move one line up }
        END;
        h := ExtCreateRegion(NIL, SizeOf(TRgnDataHeader) +
          (SizeOf(TRect) * MaxRects), pData^);        { Create a region }
        IF (Rslt <> 0) THEN BEGIN                     { Region already exists }
          CombineRgn(Rslt, Rslt, h, RGN_OR);          { Combine regions }
          DeleteObject(h);                            { Delete old region }
        END ELSE Rslt := h;                           { Return this region }
        GlobalFree(hData);                            { Clean up data }
        SelectObject(MemDc, HoldObj);                 { Select original object }
        DeleteObject(hbm32);                          { Delete dib section }
      END;
      DeleteDC(MemDC);                                { Delete device context }
    END;
    DeleteObject(Pal);                                { Destroy palette }
    DeleteObject(InBmp);                              { Delete bitmap object }
  END;
  BitmapToRegion := Rslt;                             { Return the region }
END;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{  SkinHandler -> Main window handler                Updated -> 8May02 LdB  }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{  Pretty basic windows handler. On creation it loads the bitmap image to   }
{  a device context which it uses in the paint call. On destruction it then }
{  frees this bitmap. The paint message simply paints the bitmap.           }
{  The hittest is returned so it always drags and the close button command  }
{  is accepted and closes the app.                                          }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
FUNCTION SkinHandler(Wnd: HWND; iMsg, wParam, lParam: LongInt): LongInt; STDCALL;
VAR R: TRect; MemDc: HDc; BitMap: HBitmap; Bm: TBitmap; Ps: TPaintStruct;
    Pal: HPalette;
BEGIN
   SkinHandler := 0;                                  { Preset zero return }
   Case iMsg of
     WM_CREATE: BEGIN                                 { WM_CREATE MESSAGE }
       MemDC := CreateCompatibleDC(0);                { Create memory DC }
       BitMap := LoadBmpFromFile('SKINTUT2.BMP', Pal);{ Load bitmap }
       DeleteObject(Pal);                             { Destroy created palette }
       GetObject(Bitmap, SizeOf(Bm), @Bm);            { Get bitmap data }
       SelectObject(MemDC, Bitmap);                   { Select bitmap }
       SetProp(Wnd, 'DCWITHBITMAP', MemDc);           { Set memdc to property }
       MoveWindow(Wnd, 0, 0, Bm.BmWidth, Bm.BmHeight, False);
     END;
     WM_DESTROY: BEGIN                                { WM_DESTROY MESSAGE }
       MemDc := GetProp(Wnd, 'DCWITHBITMAP');         { Retreive property }
       If (MemDc <> 0) THEN DeleteObject(MemDc);      { Dispose context }
       PostQuitMessage(0);                            { Post quit message }
     END;
     WM_PAINT: BEGIN                                  { WM_PAINT MESSAGE }
       GetWindowRect(Wnd, R);
       MemDc := GetProp(Wnd, 'DCWITHBITMAP');         { Retreive property }
       BeginPaint(Wnd, Ps);                           { Get paint structure }
       BitBlt(Ps.hDc, 0, 0, R.Right-R.Left,
         R.Bottom-R.Top, MemDC, 0, 0, SRCCopy);       { Copy bitmap screen }
       EndPaint(Wnd, Ps);                             { End of painting }
     END;
     WM_NCHITTEST: BEGIN                              { WM_NCHITTEST MESSAGE }
       SkinHandler := htCaption;                      { Return caption hit }
     END;
     WM_COMMAND: Begin                                { WM_COMMAND MESSAGE }
       CASE LoWord(wParam) Of                         { Low word wparam is cmd }
         100: PostQuitMessage(0);                     { Close button id }
         ELSE Skinhandler := DefWindowProc(Wnd, iMsg,
           wParam, lParam);                           { Call default handler }
       END;
     END;
     ELSE Skinhandler := DefWindowProc(Wnd, iMsg,
       wParam, lParam);                               { Call default handler }
   END;
END;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{  Main program code start                           Updated -> 8May02 LdB  }
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
VAR Li: LongInt; Wnd: HWND; Msg: TMsg; Wndclass: TWndClassEx;
BEGIN
   If NOT GetClassInfoEx(0, szClassName, WndClass)    { Check not registered }
   THEN BEGIN
     WITH Wndclass DO BEGIN                           { With record }
       cbSize        := Sizeof(Wndclass);             { Record size }
       style         := 0;                            { No special style }
       lpfnWndProc   := @SkinHandler;                 { Handler to use }
       cbClsExtra    := 0;                            { No extra colors }
       cbWndExtra    := 0;                            { No extra data }
       hInstance     := System.hInstance;             { Set instance }
       hIcon         := LoadIcon(0, IDI_APPLICATION); { Normal icon }
       hCursor       := LoadCursor(0, IDC_ARROW) ;    { Normal cursor }
       hbrBackground := GetStockObject(Null_Brush);   { No backbround }
       lpszMenuName  := Nil;                          { No menu }
       lpszClassName := szClassName;                  { Class name }
       hIconSm       := LoadIcon(0, IDI_APPLICATION); { Normal small icon }
     End;
     RegisterClassEx(Wndclass);                       { Register the class }
   END;
   Wnd := CreateWindowEx(0, szClassName, 'SKIN DEMO',
     0, CW_USEDEFAULT,
     CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
     0, 0, hInstance, Nil) ;                          { Create skin window }

   {* we must take caption off the window *}
   Li := GetWindowLong(Wnd, gwl_style);               { Get current style }
   Li := Li AND NOT WS_CAPTION;                       { Turn caption off }
   Li := Li AND NOT WS_ThickFrame;                    { Turn frame off }
   SetWindowLong(Wnd, gwl_style, Li);                 { Set the window style }


   {* make the window the desired shape by cutting out selected colour(s) *}
   {Ref1 -> Change the RGB values to cut another colour out }
   SetWindowRgn(Wnd,
     BitmapToRegion('SKINTUT2.BMP',
     RGB($FF, 0, $FF), $10),                          { Colour to cut }
     False);                                          { Set rgn from bmp to window }

   ShowWindow(Wnd, sw_ShowNormal);                    { Show window }
   UpdateWindow(Wnd);                                 { Update window }

   CreateWindowEx(0, 'BUTTON', 'EXIT',
     WS_VISIBLE OR WS_CHILD,
     140, 230, 80, 40,
     Wnd, 100, hInstance, Nil);                       { Create exit button }

   WHILE (GetMessage(Msg, 0, 0, 0)=True) DO BEGIN     { Wait for message }
     TranslateMessage(Msg);                           { Translate message }
     DispatchMessage(Msg);                            { Dispatch message }
   END;

   UnRegisterClass(szClassName, HInstance);           { Unregister class }
END.




